"""World state and behavior (wiring for state and reducers)
"""
from contextlib import contextmanager
from functools import reduce
from .action_factory import DuplicateAction
from .context import dynamic
from .export import export_state
from .state import empty, reduce_state


class Store:
    """ Represents state of the world together with its behavior

    State is changing due to time (context) and actions (entities),
    store describes how the state evolves and how it looks right now
    """
    def __init__(self, state=empty, context=dynamic, hooks=None):
        self.context = context
        self.initial_state = state + context.snapshot
        self.actions = []
        self.hooks = hooks or self.Hooks()

    class Hooks:
        """Base class providing no-op behavior for all available hooks
        """
        # more hook points expected in the future
        # pylint:disable=too-few-public-methods
        def post_commit(self, state, diff):
            pass

    @property
    def state(self):
        return reduce(reduce_state, self.actions, self.initial_state)

    def compute_diff(self):
        return compute_diff(self.initial_state, self.state)

    def add(self, action):
        try:
            action = action.at(self.state + self.context.snapshot)
        except DuplicateAction as exc:
            return exc.action
        else:
            self.actions.append(action)
            return action

    def commit(self):
        """Squash all actions to create new initial state
        """
        diff = self.compute_diff()
        self.initial_state = self.state
        self.actions = []
        self.hooks.post_commit(state=self.state, diff=diff)

    def export(self, dirpath):
        export_state(dirpath=dirpath, state=self.state)


    @classmethod
    @contextmanager
    def open(cls, *args, **kwargs):
        store = cls(*args, **kwargs)
        yield store
        store.commit()


def compute_diff(old_state, new_state):
    return compute_entities_diff(old_state.entities, new_state.entities)


def compute_entities_diff(old_entities, new_entities):
    """ Return list of changes between the old and new entities
    """
    # TODO: refactor to make it more readable and easily testable
    diff = []
    for entity_class, entity_map in new_entities.items():
        if entity_map.original_entities is old_entities[entity_class]:
            diff.extend(
                (entity_class, entity_id, new_entity)
                for entity_id, new_entity in entity_map.modified_entities.items()
            )
        else:
            diff.extend(
                (entity_class, entity_id, entity_map.get(entity_id, None))
                for entity_id in entity_map.keys() | old_entities[entity_class].keys()
                if entity_map.get(entity_id, None) \
                   is not old_entities[entity_class].get(entity_id, None)
            )
    return diff



""" All available blocks
"""
from ..entities import Block

blocks = (
    Block(block_id='fly', order=1),
    Block(block_id='shoot', order=2),
    Block(block_id='repeat', order=3),
    Block(block_id='while', order=4),
    Block(block_id='color', order=5),
    Block(block_id='position', order=6),
    Block(block_id='if', order=7),
    Block(block_id='if-else', order=8),
)



""" All levels
"""
from ..entities import Level

# pylint:disable=line-too-long
levels = (
    Level(level_id=1, credits=0),
    Level(level_id=2, credits=4),
    Level(level_id=3, credits=9),
    Level(level_id=4, credits=16),
    Level(level_id=5, credits=25),
    Level(level_id=6, credits=36),
    Level(level_id=7, credits=49),
    Level(level_id=8, credits=64),
    Level(level_id=9, credits=81),
)



""" Context is a part of the world state changing (potentially) continuously
"""
from datetime import datetime, timedelta
from itertools import count
from random import randint
from uuid import uuid4
from pyrsistent import PClass, field
from flocs import __version__


class Context(PClass):
    """ Base class providing context like time and randomness seed

    Attributes:
        - version: flocs version string
        - time: current datetime
        - randomness: seed for randomized algorithms
        - new_id: number or 'uuid' or 'seq'
    """
    default_time = datetime(1, 1, 1)

    version = __version__
    time = field(initial=default_time)
    randomness = field(initial=0)
    new_id = field(initial=0)

    @property
    def snapshot(self):
        return self


class DynamicContext:
    """ Dynamic context
    """
    def __init__(self):
        self.version = __version__
        self.new_id = 'uuid'

    @property
    def time(self):
        return datetime.now()

    @property
    def randomness(self):
        return randint(a=0, b=2**30)

    @property
    def snapshot(self):
        return Context(time=self.time, randomness=self.randomness, new_id=self.new_id)


class SimulationContext:
    """ Mutable simulation context
    """
    def __init__(self, time=datetime(1, 1, 1, 0, 0, 0), randomness=0, new_id='seq'):
        self.version = __version__
        self.time = time
        self.randomness = randomness
        self.new_id = new_id

    @property
    def snapshot(self):
        return Context(time=self.time, randomness=self.randomness, new_id=self.new_id)


static = Context()
dynamic = DynamicContext()


#def is_context(instance):
#    return all(hasattr(instance, attr) for attr in ['version', 'time', 'randomness', 'new_id'])


#def static(time=Context.default_time, randomness=0, new_id='seq'):
#    return Context(time=time, randomness=randomness, new_id=new_id)


#def dynamic():
#    context = Context(
#        time=datetime.now(),
#        randomness=randint(a=0, b=2**30),
#        new_id='uuid',
#    )
#    return context


def generate_id_if_not_set(maybe_id, generator=uuid4):
    just_id = maybe_id if maybe_id is not None else generator()
    return just_id



""" Conversion between various naming conventions
"""
from functools import singledispatch
import re

@singledispatch
def camel_to_snake_case(name):
    partially_underscored = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    fully_underscored = re.sub('([a-z0-9])([A-Z])', r'\1_\2', partially_underscored)
    return fully_underscored.lower()


@camel_to_snake_case.register(dict)
def _camel_to_snake_case(mapping):
    return {camel_to_snake_case(key): value for key, value in mapping.items()}


@singledispatch
def camel_to_kebab_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', name)
    kebab_case_name = re.sub('([a-z0-9])([A-Z])', r'\1-\2', s1).lower()
    return kebab_case_name


@singledispatch
def kebab_to_snake_case(name):
    return name.replace('-', '_')


@kebab_to_snake_case.register(dict)
def _kebab_to_snake_case(mapping):
    return {kebab_to_snake_case(key): value for key, value in mapping.items()}


def pluralize(name):
    if name[-1] in 'szx':
        return name + 'es'
    if name[-1] == 'y':
        return name[:-1] + 'ies'
    return name + 's'



""" Common entities for testing purposes
"""
from flocs.entities import Task, Student

s1 = Student(student_id=1, credits=0)

t1 = Task(task_id=1, category_id=None, setting=None, solution=None)
t2 = Task(task_id=2, category_id=None, setting=None, solution=None)
t3 = Task(task_id=3, category_id=None, setting=None, solution=None)
t4 = Task(task_id=4, category_id=None, setting=None, solution=None)
t5 = Task(task_id=5, category_id=None, setting=None, solution=None)
t6 = Task(task_id=6, category_id=None, setting=None, solution=None)
t7 = Task(task_id=7, category_id=None, setting=None, solution=None)
t8 = Task(task_id=8, category_id=None, setting=None, solution=None)
t9 = Task(task_id=9, category_id=None, setting=None, solution=None)



""" Unit tests for entities
"""

from datetime import datetime, timedelta
from pyrsistent import pmap
from flocs.entities import TaskSession, Action


def test_task_session_defaults():
    task_session = TaskSession(task_id=3, student_id=4)
    assert task_session.solved == False
    assert task_session.given_up == False


def test_task_session_spent_time():
    task_session = TaskSession(
        start=datetime(2017, 1, 10, 5, 30, 20),
        end=datetime(2017, 1, 10, 5, 31, 10),
    )
    assert task_session.time_spent == timedelta(seconds=50)


def test_action_as_dict():
    a = Action(action_id='a', type='t', data=pmap({'x': 7}),
               time=None, randomness=None, version=None)
    d = a._asdict()
    expected_dict = {
        'action_id': 'a', 'type': 't', 'data': {'x': 7},
        'time': None, 'randomness': None, 'version': None}
    assert d == expected_dict



import operator
from collections import ChainMap
from collections.abc import Mapping
from .utils.names import camel_to_snake_case


class EntityMap(Mapping):
    """ Collection of all entities of given type (i.e., 1 entity table)
    """
    def __init__(self, *maps, state=None, ordering=None):
        """ Initialize with provided map(s)
            Reference to state is only needed for complex filtering tasks that
            uses entities of other types
        """
        self.chain_map = ChainMap(*maps)
        self.ordering = ordering
        self.state = state

    @classmethod
    def from_list(cls, entity_list):
        return cls({get_id(entity): entity for entity in entity_list})

    def __getitem__(self, entity_id):
        return self.chain_map[entity_id]

    def __iter__(self):
        if self.ordering is None:
            return iter(self.chain_map)
        key = lambda pk: getattr(self[pk], self.ordering)
        return iter(sorted(self.chain_map, key=key))

    def __len__(self):
        return len(self.chain_map)

    def __repr__(self):
        return 'EntityMap({data})'.format(data=self.chain_map)

    def __str__(self):
        if len(self) == 0:
            return 'Empty EntityMap'
        msg = '{entity_class} entities:\n{entities}'.format(
            entity_class=self.entity_class.__name__,
            entities='\n'.join('* ' + str(e) for e in self.values()))
        return msg

    @property
    def entity_class(self):
        try:
            entity = next(iter(self.values()))
            return entity.__class__
        except StopIteration:
            return None

    def set(self, entity):
        """ Return a new map from self and the given entity
        It effectively updates entity if exists, creates otherwise.
        """
        entity_id = get_id(entity)
        return EntityMap({entity_id: entity}, *self.chain_map.maps)

    @property
    def original_entities(self):
        # if there was no modification, return self - this allows to perform
        # diffing in constant time by comparing references
        return self if len(self.chain_map.maps) == 1 else self.chain_map.maps[-1]

    @property
    def modified_entities(self):
        """ Return entities which were set since the original state
        """
        return ChainMap(*self.chain_map.maps[:-1])

    def exists(self):
        """ Return True iff there is at least 1 entity in the map
        """
        return len(self) >= 1

    def filter(self, **kwargs):
        """ Return a new map filtered according to given query given by kwargs

        For full kwargs specification, see Django QuerySet's filter:
        https://docs.djangoproject.com/en/dev/ref/models/querysets/#field-lookups
        But only a subset of lookups is currently implemented, see _test_entity_single

        Note that filtering works even with many maps, but beware that it's
        iterating through all element, even if some nested map has another
        (more efficient) implementation of filter method.
        """
        # if len(self.chain_map.maps) != 1:
        #     raise ValueError('Filtering only supported for initial (unmodified) maps')
        filtered_dict = {
            entity_id: entity
            for entity_id, entity in self.items()
            if self._test_entity(entity, **kwargs)
        }
        return EntityMap(filtered_dict)

    def order_by(self, field):
        return EntityMap(self, ordering=field)

    def first(self):
        try:
            return next(iter(self.values()))
        except StopIteration:
            return None

    def last(self):
        try:
            return list(self.values())[-1]
        except IndexError:
            return None

    def _test_entity(self, entity, **kwargs):
        """ Return True if a given entity satisfies condition by **kwargs
        """
        is_true = all(
            self._test_entity_single(entity, query, right)
            for query, right in kwargs.items())
        return is_true

    def _test_entity_single(self, entity, query, right):
        query_parts = query.split('__')
        left = getattr(entity, query_parts[0])
        query_type = query_parts[1] if len(query_parts) > 1 else 'exact'
        operator = get_operator(query_type)
        return operator(left, right)


def get_operator(query_type):
    operator_mapping = {
        'exact': operator.eq,
        'lte': operator.le,
        'gte': operator.ge,
        'lt': operator.lt,
        'gt': operator.gt,
    }
    try:
        return operator_mapping[query_type]
    except:
        raise ValueError('Unknown query type: {}'.format(query_type))


# TODO: move to some utils
def get_id(entity):
    return getattr(entity, get_id_field_name(entity))


def get_id_field_name(entity):
    class_name = entity.__class__.__name__
    id_field_name = camel_to_snake_case(class_name) + '_id'
    return id_field_name



""" Unit tests for flocs.entity_map.EntityMap
"""
from collections import namedtuple
import pytest
from flocs.state import EntityMap


# simple entity class for testing and a few instances
Entity = namedtuple('Entity', ['entity_id', 'a', 'b'])
e1 = Entity(entity_id='i', a=1, b=1)
e2 = Entity(entity_id='j', a=2, b=1)
e3 = Entity(entity_id='k', a=1, b=2)

# pylint:disable=unused-argument
# pylint:disable=attribute-defined-outside-init
# pylint:disable=pointless-statement
class TestEntityMap:

    @staticmethod
    def create_entity_map(*entities):
        """ Factory method to create an instance of any class providing
            EntityMap interface you want to test
        """
        return EntityMap.from_list(entities)

    #def setup_method(self, method):
    #    pass

    def test_getitem(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        assert entity_map['j'] == e2

    def test_get_nonexistent_item(self):
        entity_map = self.create_entity_map(e1, e2)
        with pytest.raises(KeyError):
            entity_map['k']

    def test_len(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        assert len(entity_map) == 3

    def test_iter(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        assert set(iter(entity_map)) == {'i', 'j', 'k'}

    def test_dict(self):
        entity_map = self.create_entity_map(e1, e2)
        assert dict(entity_map) == {'i': e1, 'j': e2}

    def test_filter(self):
        entity_map = self.create_entity_map(e1, e2, e3).filter(a=1)
        assert entity_map == self.create_entity_map(e1, e3)
        #assert dict(entity_map) == {'i': e1, 'k': e3}

    def test_filter_empty(self):
        entity_map = self.create_entity_map(e1, e2, e3).filter(a=100)
        assert entity_map == self.create_entity_map()

    def test_filter_gte(self):
        entity_map = self.create_entity_map(e1, e2, e3).filter(a__gte=2)
        assert entity_map == self.create_entity_map(e2)

    def test_filter_conjunction(self):
        entity_map = self.create_entity_map(e1, e2, e3).filter(a=1, b=2)
        assert entity_map == self.create_entity_map(e3)

    def test_set_check_structure(self):
        initial_entity_map = self.create_entity_map(e1, e2, e3)
        e4 = Entity(entity_id='l', a=1, b=1)
        updated_entity_map = initial_entity_map.set(e4)
        # set() is supposed to return EntityMap, not a subclass
        assert isinstance(updated_entity_map, EntityMap)
        assert len(updated_entity_map.chain_map.maps) == 2
        last_map = updated_entity_map.chain_map.maps[-1]
        assert last_map == initial_entity_map

    def test_set_new(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        e4 = Entity(entity_id='l', a=1, b=1)
        em_updated = entity_map.set(e4)
        # set() is supposed to return EntityMap, not a subclass
        assert em_updated == EntityMap.from_list([e1, e2, e3, e4])

    def test_set_existing(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        e3_updated = Entity(entity_id='k', a=7, b=2)
        em_updated = entity_map.set(e3_updated)
        # set() is supposed to return EntityMap, not a subclass
        assert em_updated == EntityMap.from_list([e1, e2, e3_updated])

    def test_set_chaining(self):
        e3_updated = Entity(entity_id='k', a=7, b=2)
        e4 = Entity(entity_id='l', a=1, b=1)
        entity_map = self.create_entity_map(e1, e2, e3).set(e3_updated).set(e4)
        assert entity_map == EntityMap.from_list([e1, e2, e3_updated, e4])

    def test_entity_class_property(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        assert entity_map.entity_class == Entity

    def test_order_by(self):
        entity_map = self.create_entity_map(e2, e3).order_by('a')
        assert list(entity_map) == ['k', 'j']
        entity_map = self.create_entity_map(e2, e3).order_by('b')
        assert list(entity_map) == ['j', 'k']

    def test_exists_true(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        assert entity_map.exists()

    def test_exists_false(self):
        entity_map = self.create_entity_map()
        assert not entity_map.exists()

    def test_first(self):
        entity_map = self.create_entity_map(e2, e3).order_by('entity_id')
        assert entity_map.first() == e2

    def test_first_none(self):
        entity_map = self.create_entity_map()
        assert entity_map.first() is None

    def test_last(self):
        entity_map = self.create_entity_map(e2, e3).order_by('entity_id')
        assert entity_map.last() == e3

    def test_last_none(self):
        entity_map = self.create_entity_map()
        assert entity_map.last() is None

    def test_contains_true(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        assert 'i' in entity_map

    def test_contains_false(self):
        entity_map = self.create_entity_map(e1, e2, e3)
        assert 'a' not in entity_map



""" Unit tests for reducers
"""
# pylint: disable=no-value-for-parameter, unused-argument

from datetime import datetime
from flocs import reducers
from flocs.actions import ActionType
from flocs.actions import StartSession, StartTask, SolveTask, SeeInstruction
from flocs.actions import RunProgram, EditProgram
from flocs.context import Context
from flocs.entities import Action, Student, TaskSession, SeenInstruction, Session, ProgramSnapshot
from flocs.entity_map import EntityMap
from flocs.reducers import reducer, extract_parameters
from flocs.state import empty, State
from .fixtures_entities import s1, t2


def test_get():
    assert reducers.get(Student, ActionType.start_session) == reducers.create_student_if_new


def test_reducer_decorator_signature():
    @reducer(None, None)
    def subreducer(substate, arg1, arg2):
        pass
    assert extract_parameters(subreducer, skip=1) == ('action',)


def test_reduce_decorator_passing_data():
    @reducer(None, 't')
    def subreducer(substate, a, c, context):
        assert a == 1
        assert c == 3
        assert context.time == 10
        assert context.randomness == 20
        return substate
    action = Action(action_id=1, type='t', data={'a': 1, 'b': 2, 'c': 3},
                    time=10, randomness=20, version=30)
    em = EntityMap()
    result = subreducer(em, action)
    assert result is em


def test_extract_parameters():
    def tmp_function(students, student_id):
        pass
    fn = tmp_function
    parameters = extract_parameters(fn, skip=1)
    expected_parameters = ('student_id',)
    assert parameters == expected_parameters


def test_create_student_if_new():
    state = empty + s1
    next_state = state.reduce(StartSession(student_id=37))
    s2 = Student(student_id=37, credits=0)
    assert next_state.students == EntityMap.from_list([s1, s2])


def test_start_task_creating_task_session():
    ts1 = TaskSession(task_session_id=81, student_id=13, task_id=28)
    ts2 = TaskSession(task_session_id=14, student_id=37, task_id=67)
    state = empty + s1 + t2 + ts1 + ts2 + Context(time=7, new_id=92)
    next_state = state.reduce(StartTask(student_id=1, task_id=2))
    ts3 = TaskSession(task_session_id=92, student_id=1, task_id=2, start=7, end=7)
    expected_task_sessions = EntityMap.from_list([ts1, ts2, ts3])
    assert next_state.task_sessions == expected_task_sessions


def test_start_task_creating_first_session():
    state = empty + s1 + t2 + Context(time=datetime(1, 1, 2, 0), new_id=92)
    next_state = state.reduce(StartTask(student_id=1, task_id=2))
    session = Session(session_id=92, student_id=1,
                      start=datetime(1, 1, 2, 0), end=datetime(1, 1, 2, 0))
    expected_sessions = EntityMap.from_list([session])
    assert next_state.sessions == expected_sessions


def test_start_task_creating_new_session():
    session1 = Session(session_id=1, student_id=1, start=None, end=datetime(1, 1, 1, 0))
    state = empty + s1 + t2 + session1 + Context(time=datetime(1, 1, 2, 0), new_id=92)
    next_state = state.reduce(StartTask(student_id=1, task_id=2))
    session2 = Session(session_id=92, student_id=1,
                       start=datetime(1, 1, 2, 0), end=datetime(1, 1, 2, 0))
    expected_sessions = EntityMap.from_list([session1, session2])
    assert next_state.sessions == expected_sessions


def test_start_task_not_creating_session():
    session1 = Session(session_id=1, student_id=1, start=None, end=datetime(1, 1, 1, 0))
    state = empty + s1 + t2 + session1 + Context(time=datetime(1, 1, 1, 0), new_id=92)
    next_state = state.reduce(StartTask(student_id=1, task_id=2))
    expected_sessions = EntityMap.from_list([session1])
    assert next_state.sessions == expected_sessions


def test_solve_task_session():
    ts1 = TaskSession(task_session_id=81, student_id=13, task_id=28,
                      start=datetime(1, 1, 1, 10), end=datetime(1, 1, 1, 11))
    ts2 = TaskSession(task_session_id=14, student_id=37, task_id=67,
                      start=datetime(1, 1, 1, 12), end=datetime(1, 1, 1, 12))
    state = empty + ts1 + ts2 + Context(time=datetime(1, 1, 1, 13))
    next_state = state.reduce(SolveTask(task_session_id=14))
    ts2s = ts2._replace(solved=True, end=datetime(1, 1, 1, 13))
    assert next_state.task_sessions == EntityMap.from_list([ts1, ts2s])


def test_solve_task_updating_session():
    session = Session(session_id=1, student_id=1, start=None, end=datetime(1, 1, 1, 0))
    ts = TaskSession(task_session_id=21, student_id=1, task_id=2,
                     start=datetime(1, 1, 1, 0), end=datetime(1, 1, 1, 0))
    state = empty + s1 + t2 + ts + session + Context(time=datetime(1, 1, 1, 0, 5))
    next_state = state.reduce(SolveTask(task_session_id=21))
    updated_session = session._replace(end=datetime(1, 1, 1, 0, 5))
    expected_sessions = EntityMap.from_list([updated_session])
    assert next_state.sessions == expected_sessions


def test_run_program_updating_session():
    session = Session(session_id=1, student_id=1, start=None, end=datetime(1, 1, 1, 0))
    ts = TaskSession(task_session_id=21, student_id=1, task_id=2,
                     start=datetime(1, 1, 1, 0), end=datetime(1, 1, 1, 0))
    state = empty + s1 + t2 + ts + session + Context(time=datetime(1, 1, 1, 0, 5))
    next_state = state.reduce(RunProgram(task_session_id=21, program=None, correct=False))
    updated_session = session._replace(end=datetime(1, 1, 1, 0, 5))
    expected_sessions = EntityMap.from_list([updated_session])
    assert next_state.sessions == expected_sessions


def test_run_program_updating_task_session():
    ts1 = TaskSession(task_session_id=81, student_id=13, task_id=28,
                      start=datetime(1, 1, 1, 10), end=datetime(1, 1, 1, 11))
    ts2 = TaskSession(task_session_id=14, student_id=37, task_id=67,
                      start=datetime(1, 1, 1, 12), end=datetime(1, 1, 1, 12))
    state = empty + ts1 + ts2 + Context(time=datetime(1, 1, 1, 20))
    next_state = state.reduce(RunProgram(task_session_id=14, program=None, correct=False))
    ts2_updated = ts2._replace(end=datetime(1, 1, 1, 20))
    assert next_state.task_sessions == EntityMap.from_list([ts1, ts2_updated])


def test_see_instruction():
    si1 = SeenInstruction(seen_instruction_id=10, student_id=1, instruction_id=2)
    state = empty + si1 + Context(new_id=15)
    next_state = state.reduce(SeeInstruction(student_id=1, instruction_id=7))
    si2 = SeenInstruction(seen_instruction_id=15, student_id=1, instruction_id=7)
    assert next_state.seen_instructions == EntityMap.from_list([si1, si2])


def test_see_instruction_more_students():
    si1 = SeenInstruction(seen_instruction_id=10, student_id=1, instruction_id=2)
    state = empty + si1 + Context(new_id=15)
    next_state = state.reduce(SeeInstruction(student_id=8, instruction_id=2))
    si2 = SeenInstruction(seen_instruction_id=15, student_id=8, instruction_id=2)
    assert next_state.seen_instructions == EntityMap.from_list([si1, si2])


def test_see_same_instruction_again():
    si1 = SeenInstruction(seen_instruction_id=10, student_id=1, instruction_id=2)
    state = empty + si1 + Context(new_id=15)
    next_state = state.reduce(SeeInstruction(student_id=1, instruction_id=2))
    assert next_state.seen_instructions == EntityMap.from_list([si1])


def test_run_program_creates_snapshot():
    ts = TaskSession(task_session_id=14, student_id=37, task_id=67,
                     start=datetime(1, 1, 1, 12), end=datetime(1, 1, 1, 12))
    state = empty + ts + Context(time=datetime(1, 1, 1, 20), new_id=10)
    next_state = state.reduce(RunProgram(task_session_id=14, program='x', correct=False))
    assert next_state.program_snapshots == EntityMap.from_list([
        ProgramSnapshot(program_snapshot_id=10, task_session_id=14, order=1,
                        time=datetime(1, 1, 1, 20), program='x', execution=True, correct=False),
    ])


def test_run_program_creates_snapshot_correct():
    ts = TaskSession(task_session_id=14, student_id=37, task_id=67,
                     start=datetime(1, 1, 1, 12), end=datetime(1, 1, 1, 12))
    state = empty + ts + Context(time=datetime(1, 1, 1, 20), new_id=10)
    next_state = state.reduce(RunProgram(task_session_id=14, program='x', correct=True))
    assert next_state.program_snapshots == EntityMap.from_list([
        ProgramSnapshot(program_snapshot_id=10, task_session_id=14, order=1,
                        time=datetime(1, 1, 1, 20), program='x', execution=True, correct=True),
    ])


def test_run_program_again_creates_another_snapshot():
    ts = TaskSession(
        task_session_id=14,
        student_id=37,
        task_id=67,
        start=datetime(1, 1, 1, 12),
        end=datetime(1, 1, 1, 12))
    ps = ProgramSnapshot(
        program_snapshot_id=10,
        task_session_id=14,
        order=1,
        time=datetime(1, 1, 1, 20),
        program='x',
        execution=False,
        correct=None)
    state = State.build(
        ts, ps,
        Context(time=datetime(1, 1, 1, 20), new_id=10))
    next_state = state.reduce(RunProgram(task_session_id=14, program='y', correct=True))
    assert next_state.program_snapshots == EntityMap.from_list([
        ps,
        ProgramSnapshot(
            program_snapshot_id=10,
            task_session_id=14,
            order=2,
            time=datetime(1, 1, 1, 20),
            program='y',
            execution=True,
            correct=True),
    ])


def test_edit_program_creates_snapshot():
    ts = TaskSession(task_session_id=14, student_id=37, task_id=67,
                     start=datetime(1, 1, 1, 12), end=datetime(1, 1, 1, 12))
    state = empty + ts + Context(time=datetime(1, 1, 1, 20), new_id=10)
    next_state = state.reduce(EditProgram(task_session_id=14, program='x'))
    assert next_state.program_snapshots == EntityMap.from_list([
        ProgramSnapshot(
            program_snapshot_id=10,
            task_session_id=14,
            order=1,
            time=datetime(1, 1, 1, 20),
            program='x',
            execution=False,
            correct=None),
    ])


def test_identity_defaultdict():
    initial_dict = {'a': 1}
    processed_dict = reducers.identity_defaultdict(initial_dict)
    assert processed_dict == initial_dict
    assert processed_dict['b']('S', 'a') == 'S'


def test_identity_reducer():
    state = {'x': [10, 20]}
    assert reducers.identity_reducer(state, 'action') == {'x': [10, 20]}



""" All available categories
"""
from ..entities import Category

# pylint:disable=line-too-long
# pylint:disable=bad-whitespace
categories = (
    Category(category_id='moves',           level_id=1, toolbox_id='fly'),
    Category(category_id='world',           level_id=2, toolbox_id='shoot'),
    Category(category_id='repeat',          level_id=3, toolbox_id='repeat'),
    Category(category_id='while',           level_id=4, toolbox_id='while'),
    Category(category_id='loops',           level_id=5, toolbox_id='loops'),
    Category(category_id='if',              level_id=6, toolbox_id='loops+if'),
    Category(category_id='comparing',       level_id=7, toolbox_id='loops+if+position'),
    Category(category_id='if-else',         level_id=8, toolbox_id='loops+if+else'),
    Category(category_id='final-challenge', level_id=9, toolbox_id='complete'),
    Category(category_id='uncategorized',   level_id=9, toolbox_id='complete'),
)



""" All available instructions
"""
from ..entities import Instruction

# pylint:disable=line-too-long
# pylint:disable=bad-whitespace
instructions = (
    Instruction(instruction_id='env.space-world'),
    Instruction(instruction_id='env.toolbox'),
    Instruction(instruction_id='env.snapping'),
    Instruction(instruction_id='env.controls'),
    Instruction(instruction_id='object.asteroid'),
    Instruction(instruction_id='object.meteoroid'),
    Instruction(instruction_id='object.diamond'),
    Instruction(instruction_id='object.wormhole'),
    Instruction(instruction_id='diamonds-status'),
    Instruction(instruction_id='energy-status'),
    Instruction(instruction_id='length-limit'),
    Instruction(instruction_id='block.fly'),
    Instruction(instruction_id='block.shoot'),
    Instruction(instruction_id='block.repeat'),
    Instruction(instruction_id='block.while'),
    Instruction(instruction_id='block.color'),
    Instruction(instruction_id='block.position'),
    Instruction(instruction_id='block.if'),
    Instruction(instruction_id='block.if-else'),
)



""" Functions describing how the world changes after various actions
"""
from collections import ChainMap, defaultdict
from inspect import signature
from . import entities
from .actions import ActionType
from .entities import Session, Student, TaskSession, SeenInstruction, ProgramSnapshot


def get(entity_class, action_type):
    """ Return a reducer for given entity class and action type
    """
    return _entity_reducer_map[entity_class][action_type]


def reducer(entity_class, action_type):
    """ Change function signature to accept whole action entity

    Additionaly, it checks that the function is called for declared entity map
    and action type
    """
    def reducer_decorator(fn):
        # Don't use @wraps, because it would retract new function signature
        def fn_wrapped(entity_map, action):
            action_types = [action_type] if isinstance(action_type, ActionType) else action_type
            assert action.type in action_types
            assert entity_map.entity_class == entity_class or not entity_map.entity_class
            data_context = ChainMap(action.data, {'context': action.context})
            kwarg_names = extract_parameters(fn, skip=1)
            kwargs = {name: data_context[name] for name in kwarg_names}
            return fn(entity_map, **kwargs)
        fn_wrapped.__name__ = fn.__name__
        fn_wrapped.__doc__ = fn.__doc__
        return fn_wrapped
    return reducer_decorator


def extract_parameters(fn, skip=0):
    return tuple(signature(fn).parameters)[skip:]


@reducer(entities.Session, ActionType.start_session)
def create_session(sessions, session_id, student_id, context):
    session = Session(
        session_id=session_id,
        student_id=student_id,
        start=context.time,
        end=context.time,
    )
    return sessions.set(session)


@reducer(entities.Session, ActionType.start_task)
def create_or_update_session(sessions, session_id, student_id, context):
    if session_id in sessions:
        session = sessions[session_id]._replace(end=context.time)
    else:
        session = Session(
            session_id=session_id,
            student_id=student_id,
            start=context.time,
            end=context.time,
        )
    # TODO: dry using create_session reducer (nontrivial because its signature
    # is modified by the @reducer decorator)
    return sessions.set(session)


@reducer(entities.Session, [ActionType.solve_task, ActionType.run_program, ActionType.edit_program])
def update_session_end(sessions, session_id, context):
    if session_id not in sessions:
        return sessions
    session = sessions[session_id]._replace(end=context.time)
    return sessions.set(session)


@reducer(entities.Student, ActionType.start_session)
def create_student_if_new(students, student_id):
    student = Student(student_id=student_id, credits=0)
    return students.set(student)


@reducer(entities.TaskSession, ActionType.start_task)
def create_task_session(task_sessions, task_session_id, student_id, task_id, context):
    task_session = TaskSession(
        task_session_id=task_session_id,
        student_id=student_id,
        task_id=task_id,
        solved=False,
        given_up=False,
        start=context.time,
        end=context.time,
    )
    return task_sessions.set(task_session)


@reducer(entities.TaskSession, ActionType.solve_task)
def solve_task_session(task_sessions, task_session_id, context):
    task_session = task_sessions[task_session_id]
    if task_session.solved:
        return task_sessions
    updated_task_session = task_session._replace(solved=True, end=context.time)
    return task_sessions.set(updated_task_session)


@reducer(entities.TaskSession, [ActionType.run_program, ActionType.edit_program])
def update_task_session_end(task_sessions, task_session_id, context):
    task_session = task_sessions[task_session_id]
    if task_session.solved:
        return task_sessions
    updated_task_session = task_session._replace(end=context.time)
    return task_sessions.set(updated_task_session)


@reducer(entities.TaskSession, ActionType.give_up_task)
def give_up_task_session(task_sessions, task_session_id):
    task_session = task_sessions[task_session_id]
    updated_task_session = task_session._replace(given_up=True)
    return task_sessions.set(updated_task_session)


@reducer(entities.SeenInstruction, ActionType.see_instruction)
def create_or_update_seen_instruction(seen_instructions, seen_instruction_id,
                                      student_id, instruction_id):
    """ Create record of new seen instruction for student-instruction pair.
        If it has already existed, nothing is changed (so seen_instruction_id
        is ignored in this case).
    """
    records = seen_instructions.filter(student_id=student_id, instruction_id=instruction_id)
    already_seen = len(records) > 0
    if already_seen:
        return seen_instructions
    seen_instruction = SeenInstruction(
        seen_instruction_id=seen_instruction_id,
        student_id=student_id,
        instruction_id=instruction_id,
    )
    updated_seen_instructions = seen_instructions.set(seen_instruction)
    return updated_seen_instructions


@reducer(entities.ProgramSnapshot, ActionType.edit_program)
def take_snapshot_on_edit(snapshots, program_snapshot_id, task_session_id,
                          order, program, context):
    snapshot = ProgramSnapshot(
        program_snapshot_id=program_snapshot_id,
        task_session_id=task_session_id,
        order=order,
        time=context.time,
        program=program,
        execution=False,
        correct=None)
    return snapshots.set(snapshot)


@reducer(entities.ProgramSnapshot, ActionType.run_program)
def take_snapshot_on_execution(snapshots, program_snapshot_id, task_session_id,
                               order, program, correct, context):
    snapshot = ProgramSnapshot(
        program_snapshot_id=program_snapshot_id,
        task_session_id=task_session_id,
        order=order,
        time=context.time,
        program=program,
        execution=True,
        correct=correct)
    return snapshots.set(snapshot)


# --------------------------------------------------------------------------


def identity_defaultdict(dictionary=None):
    dictionary = dictionary or {}
    return defaultdict(lambda: identity_reducer, dictionary)


def identity_reducer(state, _action):
    return state


ALWAYS_IDENTITY = identity_defaultdict()


# it is made explitic which entities are not changing to get a guarantee that
# an entity key corresponds to an actual entity (e.g. not just a string)
_entity_reducer_map = {
    entities.Session: identity_defaultdict({
        ActionType.start_session: create_session,
        ActionType.start_task: create_or_update_session,
        ActionType.edit_program: update_session_end,
        ActionType.run_program: update_session_end,
        ActionType.solve_task: update_session_end,
    }),
    entities.Student: identity_defaultdict({
        ActionType.start_session: create_student_if_new,
    }),
    entities.SeenInstruction: identity_defaultdict({
        ActionType.see_instruction: create_or_update_seen_instruction,
    }),
    entities.TaskSession: identity_defaultdict({
        ActionType.start_task: create_task_session,
        ActionType.edit_program: update_task_session_end,
        ActionType.run_program: update_task_session_end,
        ActionType.solve_task: solve_task_session,
        ActionType.give_up_task: give_up_task_session,
    }),
    entities.ProgramSnapshot: identity_defaultdict({
        ActionType.edit_program: take_snapshot_on_edit,
        ActionType.run_program: take_snapshot_on_execution,
    }),
    entities.Block: ALWAYS_IDENTITY,
    entities.Category: ALWAYS_IDENTITY,
    entities.Instruction: ALWAYS_IDENTITY,
    entities.Level: ALWAYS_IDENTITY,
    entities.Toolbox: ALWAYS_IDENTITY,
    entities.Task: ALWAYS_IDENTITY,
}



""" Pure functions extracting information from the world state
"""
from collections import namedtuple
from datetime import timedelta
from itertools import accumulate, chain
from math import inf
from uuid import uuid4
from flocs.entities import Level
from flocs import recommendation


PracticeOverview = namedtuple('PracticeOverview', [
    'level',
    'credits',
    'active_credits',
    'instructions',
    'tasks',
    'recommendation',
])


Recommendation = namedtuple('Recommendation', [
    'available',
    'task_id',
])


StudentInstruction = namedtuple('StudentInstruction', [
    'instruction_id',
    'seen',
])


StudentTask = namedtuple('StudentTask', [
    'task_id',
    'solved',
    'time',
])


TaskStats = namedtuple('TaskStats', [
    'task_id',
    'started_count',
    'solved_count',
])


def new_id(state):
    if state.context.new_id == 'uuid':
        return uuid4()
    if state.context.new_id == 'seq':
        return state.total_entities()
    return state.context.new_id


def get_student_id_for_task_session(state, task_session_id):
    task_session = state.task_sessions[task_session_id]
    return task_session.student_id


def get_task_id_for_task_session(state, task_session_id):
    task_session = state.task_sessions[task_session_id]
    return task_session.task_id


def get_current_session_id(state, student_id, new_if_none=True):
    session = state.sessions.filter(student_id=student_id).order_by('end').last()
    if not session or session_too_old(session, state.context.time):
        return new_id(state) if new_if_none else None
    return session.session_id


def session_too_old(session, time):
    time_passed = time - session.end
    too_old = time_passed >= timedelta(hours=5)
    return too_old


def get_recommendation(state, student_id):
    task_id = recommendation.fixed_then_random(state, student_id)
    return Recommendation(available=True, task_id=task_id)


def get_practice_overview(state, student_id):
    student = state.students[student_id]
    overview = PracticeOverview(
        level=get_level(state, student_id).level_id,
        credits=student.credits,
        active_credits=get_active_credits(state, student_id),
        instructions=get_student_instructions(state, student_id),
        tasks=get_student_tasks(state, student_id),
        recommendation=get_recommendation(state, student_id),
    )
    return overview


def get_student_instructions(state, student_id):
    seen_instructions = state.seen_instructions.filter(student_id=student_id)
    # TODO: allow for student.seen_instructions
    student_instructions = [
        StudentInstruction(
            instruction_id=instruction_id,
            seen=seen_instructions.filter(instruction_id=instruction_id).exists(),
        )
        for instruction_id in state.instructions
    ]
    return student_instructions



def get_student_tasks(state, student_id):
    student_tasks = [
        StudentTask(
            task_id=task_id,
            solved=is_solved(state, student_id, task_id),
            time=get_time(state, student_id, task_id),
        )
        for task_id in state.tasks
    ]
    return student_tasks


def is_solved(state, student_id, task_id):
    solved_task_sessions = state.task_sessions.filter(student_id=student_id,
                                                      task_id=task_id,
                                                      solved=True)
    return solved_task_sessions.exists()


def get_time(state, student_id, task_id):
    """ Best time if student has solved the task, otherwise just the last time
    """
    task_sessions = state.task_sessions.filter(student_id=student_id, task_id=task_id)
    solved_task_sessions = task_sessions.filter(solved=True)
    if solved_task_sessions.exists():
        times = [ts.time_spent for ts in solved_task_sessions.values()]
        return min(times)
    last_task_session = task_sessions.order_by('end').last()
    return last_task_session.time_spent if last_task_session else None


def get_level(state, student_id):
    if not state.levels:
        return Level(level_id=0, credits=0)
    student = state.students[student_id]
    level = Level(level_id=0, credits=0)
    levels = levels_with_total_credits(state)
    for next_level, needed_credits in chain(levels, [(None, inf)]):
        if student.credits >= needed_credits:
            level = next_level
        else:
            return level


def get_active_credits(state, student_id):
    student = state.students[student_id]
    level = get_level(state, student_id)
    level_total_credits = get_level_total_credits(state, level.level_id)
    return student.credits - level_total_credits


def get_level_total_credits(state, level_id):
    for level, credits in levels_with_total_credits(state):
        if level.level_id == level_id:
            return credits
    return 0


def levels_with_total_credits(state):
    levels = state.levels.order_by('level_id').values()
    total_credits = accumulate(chain([0], levels), lambda c, l: c + l.credits)
    next(total_credits)  # omit the added 0
    return zip(levels, total_credits)


def get_task_stats(state, task_id):
    raise NotImplementedError


def get_next_snapshot_order(state, task_session_id):
    snapshots = state.program_snapshots.filter(task_session_id=task_session_id)
    return len(snapshots) + 1



""" Type definitions for domain entities
"""
from collections import namedtuple
from datetime import timedelta
from .context import Context


class Action(namedtuple('Action', ['action_id', 'type', 'data',
                                   'time', 'randomness', 'version'])):
    """ Describes an atomic event in the world we model
    """
    __slots__ = ()

    def set_context(self, context):
        action_with_context = self._replace(
            time=context.time,
            randomness=context.randomness,
            version=context.version)
        return action_with_context

    @property
    def context(self):
        return Context(time=self.time, randomness=self.randomness)

    def _asdict(self):
        d = super()._asdict()
        d['data'] = dict(d['data'])
        return d


Session = namedtuple('Session', [
    'session_id',
    'student_id',
    'start',
    'end',
])


Student = namedtuple('Student', [
    'student_id',
    'credits'
])


SeenInstruction = namedtuple('SeenInstruction', [
    'seen_instruction_id',
    'student_id',
    'instruction_id',
])


Task = namedtuple('Task', [
    'task_id',
    'category_id',
    'setting',
    'solution',
])


Category = namedtuple('Category', [
    'category_id',
    'level_id',
    'toolbox_id',
])


Level = namedtuple('Level', [
    'level_id',
    'credits',
])


Toolbox = namedtuple('Toolbox', [
    'toolbox_id',
    'block_ids',
])


Block = namedtuple('Block', [
    'block_id',
    'order',
])


class TaskSession(namedtuple('TaskSession', [
        'task_session_id', 'student_id', 'task_id',
        'solved', 'given_up',
        'start', 'end'])):
    __slots__ = ()

    def __new__(cls, task_session_id=None, student_id=None, task_id=None,
                     solved=False, given_up=False, start=None, end=None):
        return super().__new__(cls,
            task_session_id=task_session_id,
            student_id=student_id,
            task_id=task_id,
            solved=solved,
            given_up=given_up,
            start=start,
            end=end)

    @property
    def time_spent(self):
        delta = self.end - self.start
        rounded_time = delta - timedelta(microseconds=delta.microseconds)
        time = max(rounded_time, timedelta(seconds=1))
        return time


ProgramSnapshot = namedtuple('ProgramSnapshot', [
    'program_snapshot_id',
    'task_session_id',
    'order',
    'time',
    'program',
    'execution', # snapshots are taken when a program is executed (-> True) or edited (-> False)
    'correct',  # boolean for executions, None for edits
])


Instruction = namedtuple('Instruction', [
    'instruction_id',
])


#Concept = namedtuple('Concept', [
#    'concept_id',
#    'ref',
#    'subconcepts',
#])
#
#
#ActivityRecommendation = namedtuple('ActivityRecommendation', [
#    'activity_recommendation_id',
#    'student_id',
#    'activity',
#])
#
#
#TaskRecommendation = namedtuple('TaskRecommendation', [
#    'task_recommendation_id',
#    'student_id',
#    'task_id',
#])



""" Unit and integration tests for flocs.store
"""
from datetime import datetime, timedelta
import pytest
from flocs import actions
from flocs.actions import StartTask
from flocs.context import static, SimulationContext
from flocs.entities import Action, Student, Task, TaskSession, Session
from flocs.state import empty
from flocs.store import Store, compute_entities_diff
from .fixtures_entities import s1, t1, t2, t3


@pytest.fixture
def mock_hooks():
    class MockHooks(Store.Hooks):
        """Fake persistence by storing changes as instance attributes
        """
        # pylint:disable=too-few-public-methods
        def __init__(self):
            self.calls = []
        def post_commit(self, state, diff):
            self.calls.append(('post_commit', state, diff))
    return MockHooks()


def test_store_hooks_called(mock_hooks):
    with Store.open(state=empty, context=static, hooks=mock_hooks):
        pass
    expected_hook_calls = [('post_commit', empty + static, [])]
    assert mock_hooks.calls == expected_hook_calls


def test_context_is_changing():
    simulation_context = SimulationContext(time=datetime(1, 1, 1, 0, 0, 0))
    store = Store(context=simulation_context)
    assert store.state.context.time == datetime(1, 1, 1, 0, 0, 0)
    store.context.time += timedelta(minutes=20)
    store.add(actions.empty)
    assert store.state.context.time == datetime(1, 1, 1, 0, 20, 0)


def test_current_state():
    state = empty + s1 + t1
    store = Store(state=state, context=static)
    action = StartTask(student_id=1, task_id=1)
    store.add(action)
    expected_state = state.reduce(action)
    assert store.state == expected_state


def test_commmit():
    store = Store(state=empty + s1 + t1, context=static)
    action = StartTask(student_id=1, task_id=1)
    store.add(action)
    store.commit()
    expected_state = (empty + s1 + t1).reduce(action)
    assert store.actions == []
    assert store.state == expected_state


def test_store_integration(mock_hooks):
    """ Test integretion of store, state, context and actions
    """
    with Store.open(state=empty + s1 + t1, context=static, hooks=mock_hooks) as store:
        action = store.add(StartTask(student_id=1, task_id=1))
    expected_state = (empty + s1 + t1).reduce(action)
    expected_diff = [
        (Action, 0, action),
        (TaskSession, 0,
         TaskSession(task_session_id=0, student_id=1, task_id=1,
                     start=static.time, end=static.time)),
        (Session, 0,
         Session(session_id=0, student_id=1, start=static.time, end=static.time)),
    ]
    assert len(mock_hooks.calls) == 1
    assert mock_hooks.calls[0][0] == 'post_commit'
    stored_state, stored_diff = mock_hooks.calls[0][1:]
    assert stored_state == expected_state
    assert set(stored_diff) == set(expected_diff)


def test_compute_entities_diff():
    old = empty + s1 + t1
    s1_b = s1._replace(credits=100)
    new = empty + s1_b + t1 + t2 + t3
    diff = compute_entities_diff(old.entities, new.entities)
    assert set(diff) == {(Student, 1, s1_b), (Task, 2, t2), (Task, 3, t3)}


def test_reducing_action_without_optional_parameters():
    """Covers issue #52
    See https://github.com/adaptive-learning/flocs-core/issues/52
    """
    with Store.open(state=empty) as store:
        action = actions.create(type='start-session', data={})
        store.add(action)
    student = list(store.state.students.values())[0]
    assert student.student_id is not None
    assert student.credits == 0



""" Unit tests for flocs.actions
"""
from datetime import datetime
import pytest
from flocs import actions
from flocs.action_factory import DuplicateAction
from flocs.actions import StartSession, StartTask, SolveTask
from flocs.context import static, Context
from flocs.entities import Action, Session, TaskSession
from flocs.state import empty
from flocs import __version__
from .fixtures_entities import s1, t2


def test_create():
    action = actions.create(type='nothing-happens', data={})
    expected_action = actions.NothingHappens()
    assert action == expected_action


def test_create_with_data():
    action = actions.create(type='start-task', data={'student-id': 145, 'task-id': 782})
    expected_action = actions.StartTask(student_id=145, task_id=782)
    assert action == expected_action


def test_create_nothing_happens_action():
    action = actions.NothingHappens().at(empty + static)
    expected_action = Action(
        action_id=0,
        type='nothing-happens',
        data={},
        time=static.default_time,
        randomness=0,
        version=__version__,
    )
    assert action == expected_action


def test_create_start_session_action():
    action = actions.StartSession().at(empty + static)
    expected_action = Action(
        action_id=0,
        type='start-session',
        data={'session_id': 0, 'student_id': 0},
        time=static.default_time,
        randomness=0,
        version=__version__,
    )
    assert action == expected_action


def test_discarding_start_session_action():
    session = Session(session_id=10, student_id=1, start=None, end=datetime(1, 1, 1, 0))
    state = empty + session + Context(time=datetime(1, 1, 1, 0))
    with pytest.raises(DuplicateAction) as excinfo:
        StartSession(student_id=1).at(state)
    assert excinfo.value.action.data == {'session_id': 10, 'student_id': 1}


def test_not_discarding_start_session_action():
    session1 = Session(session_id=1, student_id=1, start=None, end=datetime(1, 1, 1, 0))
    session2 = Session(session_id=2, student_id=2, start=None, end=datetime(1, 1, 2, 0))
    state = empty + session1 + session2 + Context(time=datetime(1, 1, 2, 0), new_id=10)
    action = StartSession(student_id=1).at(state)
    assert action.data == {'session_id': 10, 'student_id': 1}


def test_create_start_task_action():
    action = actions.StartTask(student_id=145, task_id=782).at(empty + static)
    expected_action = Action(
        action_id=0,
        type='start-task',
        data={'student_id': 145, 'task_id': 782, 'task_session_id': 0, 'session_id': 0},
        time=static.default_time,
        randomness=0,
        version=__version__,
    )
    assert action == expected_action


def test_discarding_start_task_action():
    """ StartTask should be discarded if the task was already started in the
        current session
    """
    state = empty + s1 + t2 + Context(time=datetime(1, 1, 1), new_id=7)
    state = state.reduce(StartTask(student_id=1, task_id=2))
    state += Context(time=datetime(1, 1, 1, 0, 1), new_id=14)
    with pytest.raises(DuplicateAction) as excinfo:
        StartTask(student_id=1, task_id=2).at(state)
    expected_data = {'student_id': 1, 'task_id': 2, 'task_session_id': 7, 'session_id': 7}
    assert excinfo.value.action.data == expected_data


def test_not_discarding_start_task_action():
    state = empty + s1 + t2 + Context(time=datetime(1, 1, 1))
    state = state.reduce(StartTask(student_id=1, task_id=2))
    state += Context(time=datetime(1, 1, 2), new_id=3)
    action = StartTask(student_id=1, task_id=2).at(state)
    assert action.data == {'student_id': 1, 'task_id': 2, 'task_session_id': 3, 'session_id': 3}


def test_create_solve_task_action():
    session = Session(session_id=58, student_id=22, start=0, end=datetime(1, 1, 1, 0))
    ts = TaskSession(task_session_id=35, student_id=22, task_id=2, start=0,
                     end=datetime(1, 1, 1, 0))
    state = empty + session + ts + Context(time=datetime(1, 1, 1, 0), new_id=4)
    action = SolveTask(task_session_id=35).at(state)
    expected_action = Action(
        action_id=4,
        type='solve-task',
        data={'task_session_id': 35, 'student_id': 22, 'task_id': 2, 'session_id': 58},
        time=datetime(1, 1, 1, 0),
        randomness=0,
        version=__version__,
    )
    assert action == expected_action


def test_create_see_instruction_action():
    action = actions.SeeInstruction(student_id=745, instruction_id=23).at(empty + static)
    expected_action = Action(
        action_id=0,
        type='see-instruction',
        data={'seen_instruction_id': 0, 'student_id': 745, 'instruction_id': 23},
        time=static.default_time,
        randomness=0,
        version=__version__,
    )
    assert action == expected_action


def test_create_nonexistent_action():
    with pytest.raises(ValueError):
        actions.create(type='invalid-action-type', data={})


def test_create_action_missing_data():
    with pytest.raises(ValueError):
        actions.create(type='solve-task', data={})


def test_create_action_redundant_data():
    with pytest.raises(ValueError):
        actions.create(type='solve-task', data={'task-session-id': 34, 'redundant-field': 20})


def test_create_action_with_autofields_provided():
    action = actions.create(type='start-session', data={'student-id': 22}).at(empty + static)
    expected_action = Action(
        action_id=0,
        type='start-session',
        data={'session_id': 0, 'student_id': 22},
        time=static.default_time,
        randomness=0,
        version=__version__,
    )
    assert action == expected_action



""" Unit tests for flocs.state
"""
from collections import namedtuple
from flocs import actions
from flocs.context import Context
from flocs.entities import Action, TaskSession
from flocs.entity_map import EntityMap
from flocs.state import empty, State, reduce_entity_map, reduce_state


# simple entity class for testing and a few instances
Entity = namedtuple('Entity', ['entity_id', 'a'])
e1 = Entity(entity_id='i', a=1)
e2 = Entity(entity_id='j', a=2)
e3 = Entity(entity_id='k', a=3)


def test_add_context():
    state = empty + Context(time=1, randomness=2)
    assert state.context.time == 1
    assert state.context.randomness == 2


def test_add_entity():
    state = empty + e1
    assert set(state.entities[Entity].values()) == {e1}


def test_add_multiple_entities():
    state = empty + e1 + e2
    assert set(state.entities[Entity].values()) == {e1, e2}


def test_add_entities_dont_modify_original_state():
    state_1 = empty + e1
    state_2 = state_1 + e2
    assert Entity not in empty.entities
    assert set(state_1.entities[Entity].values()) == {e1}
    assert set(state_2.entities[Entity].values()) == {e1, e2}


#def test_add_action():
#    a1 = actions.create(type='start-session', data={'student-id': 20})
#    state = empty + a1
#    assert Entity not in empty.entities
#    assert set(state.actions.values()) == {a1}
#    students = list(state.students.values())
#    assert len(students) == 1
#    assert students[0].student_id == 20


def test_add():
    state1 = empty + e1
    state2 = state1.add([e2, e3])
    assert set(state2.entities[Entity].values()) == {e1, e2, e3}


def test_build():
    state = State.build(e1, e2)
    assert set(state.entities[Entity].values()) == {e1, e2}


#def test_add_state():
#    state_a = State(entities=pmap({1: [11, 12]}))
#    state_b = State(entities=pmap({1: [11, 13], 2: [25, 26]}))
#    state_c = state_a + state_b
#    assert state_a == State(entities=pmap({1: [11, 12]}))
#    assert state_b == State(entities=pmap({1: [11, 13], 2: [25, 26]}))
#    assert state_c == State(entities=pmap({1: [11, 12, 13], 2: [25, 26]}))


def test_reduce_entity_map():
    task_sessions = EntityMap.from_list([
        TaskSession(task_session_id=14, student_id=37, task_id=67)
    ])
    action = Action(action_id=0, type='solve-task', data={'task_session_id': 14},
                    time=10, randomness=0, version=0)
    next_entity_map = reduce_entity_map(TaskSession, task_sessions, action)
    expected_entity_map = EntityMap.from_list([
        TaskSession(task_session_id=14, student_id=37, task_id=67, end=10, solved=True)
    ])
    assert next_entity_map == expected_entity_map


def test_reduce_state():
    a1 = actions.StartSession(student_id=20)
    next_state = reduce_state(empty, a1)
    assert set(next_state.actions.values()) == {a1.at(empty)}
    students = list(next_state.students.values())
    assert len(students) == 1
    assert students[0].student_id == 20



from collections import ChainMap
from pyrsistent import pmap
from flocs import entities
from flocs.utils.names import kebab_to_snake_case, camel_to_kebab_case
from flocs.extractors import new_id


class ActionIntent:
    """ Intent which can be transformed to an Action calling `at` method
    """
    required_fields = []
    auto_fields = []

    def __init__(self, **data):
        self.data = kebab_to_snake_case(data)
        self.validate_data()

    @classmethod
    def get_type_name(cls):
        return camel_to_kebab_case(cls.__name__)

    def complete_data(self, state):
        enriched_data = {**self.data}
        for field_name, extractor, *args in self.auto_fields:
            if field_name not in enriched_data:
                # TODO: inject additional required arguments to the extractor
                arg_values = [enriched_data[arg] for arg in args]
                enriched_data[field_name] = extractor(state, *arg_values)
        self.data = enriched_data

    def validate_data(self):
        for field in self.required_fields:
            if field not in self.data:
                tpl = 'Missing field {field}, which is required for {action} action'
                msg = tpl.format(field=field, action=self.get_type_name())
                raise ValueError(msg)
        all_data_fields = [field[0] for field in self.auto_fields] + list(self.required_fields)
        for field in self.data:
            if field not in all_data_fields:
                tpl = 'Unknown field {field} passed for {action} action'
                msg = tpl.format(field=field, action=self.get_type_name())
                raise ValueError(msg)

    def discard_action(self, state):
        return False

    def check_duplicate(self, state):
        """ Raise DuplicateAction exception if this intent is a duplicate

        Concrete actions can override this method and call
        self.raise_duplicate_action() if they find that this intent should
        be discarded as (near) duplicate of an already stored action
        """
        pass

    def raise_duplicate_action(self, **duplicate_data):
        # TODO: it should probably pass the original action, but current
        # application design does not support efficient action search, so
        # instead, we just pass a fake Action with possibly needed data
        action = entities.Action(
            action_id=None, time=None, randomness=None, version=None,
            type=self.get_type_name(),
            data=pmap(ChainMap(duplicate_data, self.data)),
        )
        raise DuplicateAction(action=action)


    def at(self, state):
        """ Create an Action from this intent, filling missing data from state
        """
        self.complete_data(state)
        self.check_duplicate(state)
        action = entities.Action(
            action_id=new_id(state),
            type=self.get_type_name(),
            data=pmap(self.data),
            time=state.context.time,
            randomness=state.context.randomness,
            version=state.context.version,
        )
        return action

    def __eq__(self, other):
        return self.get_type_name() == other.get_type_name() and self.data == other.data

    def __repr__(self):
        return '<{type} {data}>'.format(type=self.__class__.__name__, data=self.data)


class DuplicateAction(Exception):
    def __init__(self, action):
        super().__init__('This action is discarded as duplicate of {}'.format(action))
        self.action = action



"""Unit tests for context
"""

from datetime import datetime, timedelta
import pytest
from flocs.context import static, dynamic, Context, SimulationContext
from flocs.context import generate_id_if_not_set

class TestContext:
    context = Context()

    def test_context_interface(self):
        assert hasattr(self.context, 'version')
        assert hasattr(self.context, 'time')
        assert hasattr(self.context, 'randomness')
        assert hasattr(self.context, 'new_id')
        assert hasattr(self.context, 'snapshot')

    def test_snapshot_returns_context(self):
        assert isinstance(self.context.snapshot, Context)


class TestDynamicContext(TestContext):
    context = dynamic


class TestStaticContext(TestContext):
    context = static

    def test_fixed_randomness(self):
        assert self.context.randomness == self.context.randomness

    def test_fixed_time(self):
        assert self.context.time == self.context.time

    def test_fixed_new_id(self):
        assert self.context.new_id == self.context.new_id == 0

    def test_immutability(self):
        with pytest.raises(AttributeError):
            static.randomness = 4

    def test_snapshot(self):
        assert static.snapshot == static


class TestContextWithParameters(TestContext):
    context = Context(time=0, randomness=1, new_id=2)

    def test_fixed_time(self):
        assert self.context.time == 0
        assert self.context.time == 0

    def test_fixed_randomness(self):
        assert self.context.randomness == 1
        assert self.context.randomness == 1

    def test_fixed_new_id(self):
        assert self.context.new_id == 2
        assert self.context.new_id == 2

    def test_snapshot(self):
        assert self.context.snapshot == Context(time=0, randomness=1, new_id=2)


class TestSimulationContext(TestContext):
    context = SimulationContext(randomness=4, time=datetime(1, 2, 3, 0, 55, 0))

    def test_fixed_randomness(self):
        assert self.context.randomness == 4
        assert self.context.randomness == 4

    def test_simulation_time(self):
        t1 = self.context.time
        assert t1 == datetime(1, 2, 3, 0, 55, 0)
        self.context.time += timedelta(hours=10)
        t2 = self.context.time
        assert t2 == datetime(1, 2, 3, 10, 55, 0)


def test_generate_id_if_not_set_when_set():
    generated_id = generate_id_if_not_set(55)
    assert generated_id == 55


def test_generate_id_if_not_set_when_not_set():
    generator = lambda: 817
    generated_id = generate_id_if_not_set(None, generator=generator)
    assert generated_id == 817



""" All available toolboxes
"""
from ..entities import Toolbox

# pylint:disable=line-too-long
# pylint:disable=bad-whitespace
toolboxes = (
    Toolbox(toolbox_id='fly',               block_ids=['fly']),
    Toolbox(toolbox_id='shoot',             block_ids=['fly', 'shoot']),
    Toolbox(toolbox_id='repeat',            block_ids=['fly', 'shoot', 'repeat']),
    Toolbox(toolbox_id='while',             block_ids=['fly', 'shoot', 'while', 'color']),
    Toolbox(toolbox_id='loops',             block_ids=['fly', 'shoot', 'repeat', 'while', 'color']),
    Toolbox(toolbox_id='loops+if',          block_ids=['fly', 'shoot', 'repeat', 'while', 'color', 'if']),
    Toolbox(toolbox_id='loops+if+position', block_ids=['fly', 'shoot', 'repeat', 'while', 'color', 'position', 'if']),
    Toolbox(toolbox_id='loops+if+else',     block_ids=['fly', 'shoot', 'repeat', 'while', 'color', 'position', 'if', 'if-else']),
    Toolbox(toolbox_id='complete',          block_ids=['fly', 'shoot', 'repeat', 'while', 'color', 'position', 'if', 'if-else']),
)



""" Unit tests for extractors
"""
from datetime import datetime, timedelta
from flocs.context import Context
from flocs.extractors import get_level, get_active_credits, get_recommendation
from flocs.extractors import get_student_instructions, StudentInstruction
from flocs.extractors import get_student_tasks, StudentTask
from flocs.extractors import get_practice_overview, PracticeOverview, Recommendation
from flocs.extractors import get_current_session_id
from flocs.extractors import get_student_id_for_task_session
from flocs.extractors import get_next_snapshot_order
from flocs.state import default, empty, State
from flocs.entities import Student, Instruction, SeenInstruction, TaskSession, Session, ProgramSnapshot
from .fixtures_entities import s1, t2, t3


def test_get_recommendation():
    state = default + s1
    recommendation = get_recommendation(state, student_id=1)
    assert recommendation.available
    assert recommendation.task_id == 'one-step-forward'


def test_get_student_tasks_no_task_sessions():
    state = empty + s1 + t2 + t3
    student_tasks = get_student_tasks(state, student_id=1)
    expected_student_tasks = {
        StudentTask(task_id=2, solved=False, time=None),
        StudentTask(task_id=3, solved=False, time=None),
    }
    assert set(student_tasks) == expected_student_tasks


def test_get_student_tasks_with_solved_task_session():
    state = empty + s1 + t2 + t3 \
        + TaskSession(task_session_id=1, student_id=1, task_id=2, solved=True,
                      start=datetime(1, 1, 1, 0, 0, 30), end=datetime(1, 1, 1, 0, 0, 40))
    student_tasks = get_student_tasks(state, student_id=1)
    expected_student_tasks = {
        StudentTask(task_id=2, solved=True, time=timedelta(seconds=10)),
        StudentTask(task_id=3, solved=False, time=None),
    }
    assert set(student_tasks) == expected_student_tasks


def test_get_student_tasks_with_unsolved_task_session():
    state = empty + s1 + t2 + t3 \
        + TaskSession(task_session_id=1, student_id=1, task_id=2,
                      start=datetime(1, 1, 1, 0, 0, 30), end=datetime(1, 1, 1, 0, 0, 40))
    student_tasks = get_student_tasks(state, student_id=1)
    expected_student_tasks = {
        StudentTask(task_id=2, solved=False, time=timedelta(seconds=10)),
        StudentTask(task_id=3, solved=False, time=None),
    }
    assert set(student_tasks) == expected_student_tasks


def test_get_student_tasks_best_solved_time():
    state = State.build(
        s1, t2, t3,
        TaskSession(task_session_id=1, student_id=1, task_id=2, solved=True,
            start=datetime(1, 1, 1, 0, 0, 10), end=datetime(1, 1, 1, 0, 0, 18)),
        TaskSession(task_session_id=2, student_id=1, task_id=2, solved=True,
            start=datetime(1, 1, 1, 0, 0, 20), end=datetime(1, 1, 1, 0, 0, 25)),
        TaskSession(task_session_id=3, student_id=1, task_id=2, solved=True,
            start=datetime(1, 1, 1, 0, 0, 30), end=datetime(1, 1, 1, 0, 0, 39)),
        TaskSession(task_session_id=3, student_id=1, task_id=2, solved=False,
            start=datetime(1, 1, 1, 0, 0, 40), end=datetime(1, 1, 1, 0, 0, 41)),
    )
    student_tasks = get_student_tasks(state, student_id=1)
    expected_student_tasks = {
        StudentTask(task_id=2, solved=True, time=timedelta(seconds=5)),
        StudentTask(task_id=3, solved=False, time=None),
    }
    assert set(student_tasks) == expected_student_tasks


def test_get_student_tasks_best_solved_datetime():
    state = State.build(
        s1, t2,
        TaskSession(task_session_id=1, student_id=1, task_id=2, solved=True,
            start=datetime(2017, 1, 10, 5, 30, 10),
            end=datetime(2017, 1, 10, 5, 30, 17)),
        TaskSession(task_session_id=1, student_id=1, task_id=2, solved=True,
            start=datetime(2017, 1, 10, 5, 30, 20),
            end=datetime(2017, 1, 10, 5, 30, 23))
    )
    student_tasks = get_student_tasks(state, student_id=1)
    expected_student_tasks = {
        StudentTask(task_id=2, solved=True, time=timedelta(seconds=3)),
    }
    assert set(student_tasks) == expected_student_tasks


def test_get_student_tasks_last_unsolved_time():
    state = State.build(
        s1, t2, t3,
        TaskSession(task_session_id=1, student_id=1, task_id=2, solved=False, start=datetime(1, 1, 1, 0, 0, 10), end=datetime(1, 1, 1, 0, 0, 18)),
        TaskSession(task_session_id=2, student_id=1, task_id=2, solved=False, start=datetime(1, 1, 1, 0, 0, 20), end=datetime(1, 1, 1, 0, 0, 25)),
        TaskSession(task_session_id=3, student_id=1, task_id=2, solved=False, start=datetime(1, 1, 1, 0, 0, 30), end=datetime(1, 1, 1, 0, 0, 32)),
    )
    student_tasks = get_student_tasks(state, student_id=1)
    expected_student_tasks = {
        StudentTask(task_id=2, solved=False, time=timedelta(seconds=2)),
        StudentTask(task_id=3, solved=False, time=None),
    }
    assert set(student_tasks) == expected_student_tasks


def test_get_student_instructions():
    iA = Instruction(instruction_id='A')
    iB = Instruction(instruction_id='B')
    state = empty + s1 + iA + iB
    state += SeenInstruction(seen_instruction_id=2, student_id=1, instruction_id='B')
    student_instructions = get_student_instructions(state, student_id=1)
    expected_student_instructions = {
        StudentInstruction(instruction_id='A', seen=False),
        StudentInstruction(instruction_id='B', seen=True),
    }
    assert set(student_instructions) == expected_student_instructions


def test_get_student_instructions_all_unseen():
    iA = Instruction(instruction_id='A')
    iB = Instruction(instruction_id='B')
    state = empty + s1 + iA + iB
    student_instructions = get_student_instructions(state, student_id=1)
    expected_student_instructions = {
        StudentInstruction(instruction_id='A', seen=False),
        StudentInstruction(instruction_id='B', seen=False),
    }
    assert set(student_instructions) == expected_student_instructions


def test_get_student_instructions_all_seen():
    iA = Instruction(instruction_id='A')
    iB = Instruction(instruction_id='B')
    state = empty + s1 + iA + iB
    state += SeenInstruction(seen_instruction_id=1, student_id=1, instruction_id='A')
    state += SeenInstruction(seen_instruction_id=2, student_id=1, instruction_id='B')
    student_instructions = get_student_instructions(state, student_id=1)
    expected_student_instructions = {
        StudentInstruction(instruction_id='A', seen=True),
        StudentInstruction(instruction_id='B', seen=True),
    }
    assert set(student_instructions) == expected_student_instructions


def test_get_level():
    student = Student(student_id=1, credits=20)
    state = default + student
    level = get_level(state, student_id=1)
    assert level.level_id == 3


def test_get_active_credits():
    student = Student(student_id=1, credits=20)
    state = default + student
    assert get_active_credits(state, student_id=1) == 7


def test_get_practice_overview_empty():
    state = empty + s1
    overview = get_practice_overview(state, student_id=1)
    expected_overview = PracticeOverview(
        level=0,
        credits=0,
        active_credits=0,
        instructions=[],
        tasks=[],
        recommendation=Recommendation(available=True, task_id='one-step-forward')
    )
    assert overview == expected_overview


def test_get_practice_overview_level_and_credits():
    state = default + s1._replace(credits=10)
    overview = get_practice_overview(state, student_id=1)
    assert overview.level == 2
    assert overview.credits == 10
    assert overview.active_credits == 6


def test_get_practice_overview_with_instructions():
    state = State.build(s1, Instruction(instruction_id='A'))
    overview = get_practice_overview(state, student_id=1)
    assert overview.instructions == [StudentInstruction(instruction_id='A', seen=False)]


def test_get_practice_overview_with_tasks():
    ts = TaskSession(student_id=1, task_id=2, solved=True,
                     start=datetime(1, 1, 1, 0, 0, 30), end=datetime(1, 1, 1, 0, 0, 40))
    state = State.build(s1, t2, ts)
    overview = get_practice_overview(state, student_id=1)
    assert overview.tasks == [StudentTask(task_id=2, solved=True, time=timedelta(seconds=10))]


def test_get_current_session_id_single_session():
    session = Session(session_id=17, student_id=21, start=None, end=datetime(1, 1, 1, 0))
    state = empty + session + Context(time=datetime(1, 1, 1, 0))
    session_id = get_current_session_id(state, student_id=21)
    assert session_id == 17


def test_get_current_session_id_latest():
    session1 = Session(session_id=17, student_id=21, start=None, end=datetime(1, 1, 1, 0))
    session2 = Session(session_id=15, student_id=21, start=None, end=datetime(1, 1, 1, 9))
    state = empty + session1 + session2 + Context(time=datetime(1, 1, 1, 9))
    session_id = get_current_session_id(state, student_id=21)
    assert session_id == 15


def test_get_current_session_id_multiple_students():
    session1 = Session(session_id=17, student_id=21, start=None, end=datetime(1, 1, 1, 8, 50))
    session2 = Session(session_id=15, student_id=22, start=None, end=datetime(1, 1, 1, 9))
    state = empty + session1 + session2 + Context(time=datetime(1, 1, 1, 9))
    session_id = get_current_session_id(state, student_id=21)
    assert session_id == 17


def test_get_current_session_first_session():
    session = Session(session_id=15, student_id=22, start=None, end=datetime(1, 1, 1, 9))
    state = empty + session + Context(time=datetime(1, 1, 1, 9), new_id=17)
    session_id = get_current_session_id(state, student_id=21)
    assert session_id == 17


def test_get_current_session_old_session():
    session = Session(session_id=15, student_id=21, start=None, end=datetime(1, 1, 1, 0))
    state = empty + session + Context(time=datetime(1, 1, 3, 0), new_id=17)
    session_id = get_current_session_id(state, student_id=21)
    assert session_id == 17


def test_get_student_id_for_task_session():
    state = empty + TaskSession(task_session_id=35, student_id=22, task_id=2, start=0, end=0)
    student_id = get_student_id_for_task_session(state, 35)
    assert student_id == 22


def test_get_first_snapshot_order():
    state = State.build(
        ProgramSnapshot(program_snapshot_id=10, task_session_id=8, order=1,
                        time=None, program=None, execution=None, correct=None),
        ProgramSnapshot(program_snapshot_id=10, task_session_id=8, order=2,
                        time=None, program=None, execution=None, correct=None),
    )
    next_order = get_next_snapshot_order(state, task_session_id=7)
    assert next_order == 1


def test_get_next_snapshot_order():
    state = State.build(
        ProgramSnapshot(program_snapshot_id=10, task_session_id=7, order=1,
                        time=None, program=None, execution=None, correct=None),
        ProgramSnapshot(program_snapshot_id=11, task_session_id=7, order=2,
                        time=None, program=None, execution=None, correct=None),
        ProgramSnapshot(program_snapshot_id=12, task_session_id=8, order=1,
                        time=None, program=None, execution=None, correct=None),
        ProgramSnapshot(program_snapshot_id=13, task_session_id=8, order=2,
                        time=None, program=None, execution=None, correct=None),
        ProgramSnapshot(program_snapshot_id=14, task_session_id=8, order=3,
                        time=None, program=None, execution=None, correct=None),
        ProgramSnapshot(program_snapshot_id=15, task_session_id=8, order=4,
                        time=None, program=None, execution=None, correct=None),
    )
    next_order = get_next_snapshot_order(state, task_session_id=7)
    assert next_order == 3



""" Common world-states for testing purposes
"""
from flocs.state import empty
from flocs.context import static
from .fixtures_entities import s1, t1, t2, t3, t4, t5

w0 = empty + static
w1 = w0 + s1 + t1 + t2 + t3 + t4 + t5



from setuptools import setup


def main():
    setup(
        name='flocs',
        version=extract_version(),
        description='Components for adaptive learning of computer science',
        author='Tomas Effenberger',
        author_email='tomas.effenberger@gmail.com',
        url='https://github.com/adaptive-learning/flocs-core',
        packages=[
            'flocs',
            'flocs.data',
            'flocs.utils',
            'flocs.tests',
        ],
        install_requires=[
            'pyrsistent==0.12.1',
        ],
        license='MIT',
    )


def extract_version():
    with open('flocs/_version.py') as version_file:
        return version_file.read().split('"')[1]


if __name__ == "__main__":
    main()



""" Techniques for task recommendation

Task recommender protocol:
    - args: state, student_id
    - return: recommneded task_id
"""
from collections import namedtuple
import random


Criterion = namedtuple('Criterion', ['weight', 'fn'])

default_fixed_order = (
    'one-step-forward',
    'diamond-on-right',
    'wormhole-demo',
    'shot',
    'ladder',
)


def randomly(state, student_id):
    """ Select task completely at random
    """
    del student_id  # intentionally unused argument
    random.seed(state.context.randomness)
    task_ids = list(state.tasks.keys())
    selected_task_id = random.choice(task_ids)
    return selected_task_id


def multicriteria(state, student_id, criteria):
    """ Recommend task linearly combining several criteria

    Usage:
    ```
    level_date_recommender = partial(
        multicriteria,
        criteria=[
            Criterion(weight=1, fn=level_difference)
            Criterion(weight=5, fn=time_from_last_solution)
        ]
    )
    ```
    """
    score_task = lambda task_id: sum(
        criterion.weight * criterion.fn(state, student_id, task_id)
        for criterion in criteria)
    best_task_id = max(state.tasks, key=score_task)
    return best_task_id


def fixed_order(state, student_id, order=default_fixed_order):
    """ Recommend task in a given fixed order

    It only moves to next task in the order when the current task is solved.
    (In particular, it's not enough to start solving the task. Giving up
    doesn't convice the recommender to move to the next task neither.)

    Additional args:
        - order: list of task IDs
    """
    solved_task_sessions = state.task_sessions.filter(
        student_id=student_id,
        solved=True)
    solved_task_ids = {ts.task_id for ts in solved_task_sessions.values()}
    for task_id in order:
        if task_id not in solved_task_ids:
            return task_id
    raise ValueError('last task reached, there is no next task')


def fixed_then_random(state, student_id):
    """ Select task in fixed order, then from other unsolved, then from all

    It never selects the previously-solved task
    """
    try:
        return fixed_order(state, student_id)
    except ValueError:
        task_ids = list(state.tasks.keys())
        solved_task_sessions = state.task_sessions.filter(
            student_id=student_id,
            solved=True)
        solved_task_ids = [ts.task_id for ts in solved_task_sessions.values()]
        unsolved_task_ids = [t_id for t_id in task_ids if t_id not in solved_task_ids]
        random.seed(state.context.randomness)
        if unsolved_task_ids:
            selected_task_id = random.choice(unsolved_task_ids)
            return selected_task_id
        else:
            last_task_session = solved_task_sessions.order_by('end').last()
            last_task_id = last_task_session.task_id
            task_ids.remove(last_task_id)
            selected_task_id = random.choice(task_ids)
            return selected_task_id



from datetime import datetime
import csv
import os
from .utils.names import camel_to_kebab_case, pluralize


def export_state(dirpath, state):
    datestamp = datetime.now().strftime('%Y-%m-%d')
    # the last empty path ('') is there to make it a directory, not a file
    full_dirpath = os.path.join(dirpath, datestamp, '')
    os.makedirs(full_dirpath, exist_ok=True)
    for entity_class, entity_map in state.entities.items():
        export_entity_map(full_dirpath, entity_class, entity_map)


def export_entity_map(dirpath, entity_class, entity_map):
    file_name = pluralize(camel_to_kebab_case(entity_class.__name__)) + '.csv'
    full_path = os.path.join(dirpath, file_name)
    fields = entity_class._fields
    with open(full_path, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, quoting=csv.QUOTE_NONNUMERIC)
        writer.writerow(fields)
        for entity in entity_map.values():
            writer.writerow(entity)



""" Actions represent events and interactions in the world we want to model
"""
from enum import Enum
from flocs.action_factory import ActionIntent
from flocs.extractors import new_id, get_current_session_id
from flocs.extractors import get_student_id_for_task_session
from flocs.extractors import get_task_id_for_task_session
from flocs.extractors import get_next_snapshot_order


def create(type='nothing-happens', data=None):
    """ Create a new action of given type and with given data

    Args:
        type: one of flocs.actions.ActionTypes
        data: dictionary of data for given action type (using kebab-case keys)

    Return:
        flocs.entities.Action
    """
    action_type = ActionType(type)
    data_dict = data or {}
    action = action_type.creator(**data_dict)
    return action


class NothingHappens(ActionIntent):
    """ Nothing particular happened (but time has passed)
    """
    auto_fields = []
    required_fields = []


class StartSession(ActionIntent):
    """ Student starts interacting with the learning system
    """
    required_fields = []
    auto_fields = [
        ('session_id', new_id),
        ('student_id', new_id),
    ]

    def check_duplicate(self, state):
        """ Don't start session if another session is still active
        """
        student_id = self.data['student_id']
        current_session_id = get_current_session_id(state, student_id, new_if_none=False)
        if current_session_id is not None:
            self.raise_duplicate_action(session_id=current_session_id)


class StartTask(ActionIntent):
    """ Student starts working on a task
    """
    required_fields = [
        'student_id',
        'task_id',
    ]
    auto_fields = [
        ('task_session_id', new_id),
        ('session_id', get_current_session_id, 'student_id'),
    ]

    def check_duplicate(self, state):
        """ Don't start task if it was already started in this session
        """
        student_id = self.data['student_id']
        task_id = self.data['task_id']
        session_id = get_current_session_id(state, student_id, new_if_none=False)
        if session_id is None:
            return
        session = state.sessions[session_id]
        task_session = state.task_sessions \
            .filter(student_id=student_id, task_id=task_id).order_by('start').last()
        if task_session and task_session.start >= session.start:
            self.raise_duplicate_action(task_session_id=task_session.task_session_id)


class EditProgram(ActionIntent):
    """ Student has made a change in their program
    """
    required_fields = (
        'task_session_id',
        'program',
    )
    auto_fields = (
        ('program_snapshot_id', new_id),
        ('student_id', get_student_id_for_task_session, 'task_session_id'),
        ('task_id', get_task_id_for_task_session, 'task_session_id'),
        ('session_id', get_current_session_id, 'student_id'),
        ('order', get_next_snapshot_order, 'task_session_id'),
    )


class RunProgram(ActionIntent):
    """ Student has executed their program
    """
    required_fields = (
        'task_session_id',
        'program',
        'correct',
    )
    auto_fields = (
        ('program_snapshot_id', new_id),
        ('student_id', get_student_id_for_task_session, 'task_session_id'),
        ('task_id', get_task_id_for_task_session, 'task_session_id'),
        ('session_id', get_current_session_id, 'student_id'),
        ('order', get_next_snapshot_order, 'task_session_id'),
    )


class SolveTask(ActionIntent):
    """ Student has solved a task
    """
    required_fields = ['task_session_id']
    auto_fields = [
        ('student_id', get_student_id_for_task_session, 'task_session_id'),
        ('task_id', get_task_id_for_task_session, 'task_session_id'),
        ('session_id', get_current_session_id, 'student_id'),
    ]


class GiveUpTask(ActionIntent):
    """ Student has given up a task
    """
    required_fields = ['task_session_id']
    auto_fields = []


class SeeInstruction(ActionIntent):
    """ Student has seen an instruction
    """
    required_fields = ['student_id', 'instruction_id']
    auto_fields = [
        ('seen_instruction_id', new_id)
    ]


class ActionType(str, Enum):
    """ Available action types and associated action creators
    """
    nothing_happens = 'nothing-happens'
    start_session = 'start-session'
    start_task = 'start-task'
    run_program = 'run-program'
    edit_program = 'edit-program'
    solve_task = 'solve-task'
    give_up_task = 'give-up-task'
    see_instruction = 'see-instruction'

    @property
    def creator(self):
        return {
            'nothing-happens': NothingHappens,
            'start-session': StartSession,
            'start-task': StartTask,
            'run-program': RunProgram,
            'edit-program': EditProgram,
            'solve-task': SolveTask,
            'give-up-task': GiveUpTask,
            'see-instruction': SeeInstruction,
        }[self.value]


empty = create(type='nothing-happens', data={})



""" Representation of a world state

State composition examples:
    - state.empty + Student(...) + Action(...)
    - state.default + SimulationContext(...) + students
"""
from collections import namedtuple, Iterable
from pyrsistent import pmap
from .action_factory import ActionIntent
from .context import static, Context
from .data import blocks, instructions, levels, toolboxes, categories, tasks
from .entities import Block, Instruction, Level, Toolbox, Category, Task
from .entities import Action, Student, TaskSession, SeenInstruction, Session, ProgramSnapshot
from .entity_map import EntityMap
from . import reducers


class State(namedtuple('State', ['entities', 'context'])):
    """ State of the world at a specific moment (immutable snapshot)
    """
    __slots__ = ()

    def __new__(cls, entities=None, context=static):
        entities = pmap(entities)
        return super().__new__(cls, entities=entities, context=context)

    @classmethod
    def build(cls, *components):
        """ Create new state from a sequence of components

        Component can be a state, entity, action, context or even iterable of
        nested components, they will be handled differently according to their
        type, see __add__ for details. (builder design patter)
        """
        return empty.add(*components)

    def add(self, *components):
        """ Create new state from a the current one, adding new components
        """
        return sum(components, self)

    def __add__(self, component):
        """ Create a new state from the current one by adding a new component

        Component can be a state, entity, action, context or even iterable of
        nested components.
        """
        #if isinstance(component, DynamicContext):
        #    return self.add_context(context=component.snapshot)
        if isinstance(component, Context):
            return self.add_context(context=component)
        #if isinstance(component, State):
        #    return self.add_state(state=component)
        #if isinstance(component, Action):
        #    return self.reduce(action=component)
        if isinstance(component, tuple):  # TODO: test for Entity base class
            return self.add_entity(entity=component)
        if isinstance(component, Iterable):
            return self.add(*component)
        raise ValueError('Cannot add {c} to the state'.format(c=component))

    #def add_state(self, state):
    #    TODO: merged_entities = self.entities.update_with(lambda l, r: ??, state.entities)
    #    new_state = state._replace(entities=merged_entities)
    #    return new_state

    def add_context(self, context):
        return self._replace(context=context)

    def add_entity(self, entity):
        if entity.__class__ in self.entities:
            old_entity_map = self.entities[entity.__class__]
            new_entity_map = old_entity_map.set(entity)
        else:
            new_entity_map = EntityMap.from_list([entity])
        new_entities = self.entities.set(entity.__class__, new_entity_map)
        new_state = self._replace(entities=new_entities)
        return new_state

    def reduce(self, action):
        # TODO: allow to reduce multiple actions at once
        return reduce_state(state=self, action=action)

    def total_entities(self):
        return sum(len(entity_map) for entity_map in self.entities.values())

    @property
    def actions(self):
        return self.entities[Action]

    @property
    def levels(self):
        return self.entities[Level]

    @property
    def instructions(self):
        return self.entities[Instruction]

    @property
    def program_snapshots(self):
        return self.entities[ProgramSnapshot]

    @property
    def seen_instructions(self):
        return self.entities[SeenInstruction]

    @property
    def sessions(self):
        return self.entities[Session]

    @property
    def students(self):
        return self.entities[Student]

    @property
    def tasks(self):
        return self.entities[Task]

    @property
    def task_sessions(self):
        return self.entities[TaskSession]

    @property
    def toolboxes(self):
        return self.entities[Toolbox]


def reduce_state(state, action):
    if isinstance(action, ActionIntent):
        action = action.at(state)
    new_entities = reduce_entities(state.entities, action)
    new_state = State(entities=new_entities, context=action.context)
    return new_state


def reduce_entities(old_entities, action):
    new_entities = pmap({
        entity_class: reduce_entity_map(entity_class, entity_map, action)
        for entity_class, entity_map in old_entities.items()
    })
    return new_entities


def reduce_entity_map(entity_class, entity_map, action):
    if entity_class == Action:
        return entity_map.set(action)
    reducer = reducers.get(entity_class, action.type)
    next_entity_map = reducer(entity_map, action)
    return next_entity_map


# some prepared immutable states

empty = State(entities=pmap({
    Block: EntityMap(),
    Instruction: EntityMap(),
    Level: EntityMap(),
    Toolbox: EntityMap(),
    Category: EntityMap(),
    Task: EntityMap(),
    Action: EntityMap(),
    Student: EntityMap(),
    TaskSession: EntityMap(),
    SeenInstruction: EntityMap(),
    Session: EntityMap(),
    ProgramSnapshot: EntityMap(),
}), context=static)

default = State(entities=pmap({
    Block: EntityMap.from_list(blocks),
    Instruction: EntityMap.from_list(instructions),
    Level: EntityMap.from_list(levels),
    Toolbox: EntityMap.from_list(toolboxes),
    Category: EntityMap.from_list(categories),
    Task: EntityMap.from_list(tasks),
    Action: EntityMap(),
    Student: EntityMap(),
    TaskSession: EntityMap(),
    SeenInstruction: EntityMap(),
    Session: EntityMap(),
    ProgramSnapshot: EntityMap(),
}), context=static)



""" Unit tests of recommendation strategies
"""
from functools import partial
import pytest
from flocs import recommendation
from flocs.context import Context
from flocs.entities import TaskSession, Student
from flocs.state import State, empty, default
from .fixtures_entities import s1, t2, t3, t5, t9


def test_randomly():
    state_a = State.build(Context(randomness=0), s1, t2, t5, t9)
    state_b = State.build(Context(randomness=1), s1, t2, t5, t9)
    task_a = recommendation.randomly(state_a, student_id=1)
    task_b = recommendation.randomly(state_b, student_id=1)
    assert task_a in {2, 5, 9}
    assert task_b in {2, 5, 9}
    assert task_a != task_b  # if equal, change randomness seeds


def test_fixed_order():
    state = State.build(t2, t5, t9, s1)
    state += TaskSession(task_session_id=1, student_id=1, task_id=5, solved=True, given_up=False)
    select_task = partial(recommendation.fixed_order, order=[5, 9, 2])
    selected_task_id = select_task(state, student_id=1)
    assert selected_task_id == 9


def test_select_task_in_fixed_order_no_skipping():
    state = State.build(t2, t5, t9, s1)
    state += TaskSession(task_session_id=1, student_id=1, task_id=9, solved=True, given_up=False)
    select_task = partial(recommendation.fixed_order, order=[5, 9, 2])
    selected_task_id = select_task(state, student_id=1)
    assert selected_task_id == 5


def test_select_task_in_fixed_order_last_task():
    state = State.build(t2, t5, s1)
    state += TaskSession(task_session_id=1, student_id=1, task_id=2, solved=True, given_up=False)
    state += TaskSession(task_session_id=2, student_id=1, task_id=5, solved=True, given_up=False)
    select_task = partial(recommendation.fixed_order, order=[2, 5])
    with pytest.raises(ValueError):
        select_task(state, student_id=1)


def test_select_same_task_until_solved():
    state = State.build(t2, t5, s1)
    select_task = partial(recommendation.fixed_order, order=[2, 5])
    assert select_task(state, student_id=1) == 2
    state += TaskSession(task_session_id=1, student_id=1, task_id=2, solved=False, given_up=False)
    assert select_task(state, student_id=1) == 2
    state += TaskSession(task_session_id=1, student_id=1, task_id=2, solved=True, given_up=False)
    assert select_task(state, student_id=1) == 5


def test_default_fixed_order():
    state = default + s1
    state += TaskSession(task_session_id=1, student_id=1, task_id='one-step-forward',
                         solved=True, given_up=False)
    assert recommendation.fixed_order(state, student_id=1) == 'diamond-on-right'


def test_fixed_then_random():
    state = default + s1
    unsolved = {task_id for task_id in state.tasks}
    m = len(state.tasks)
    for k in range(m):
        next_task = recommendation.fixed_then_random(state, student_id=1)
        assert next_task in unsolved
        unsolved.discard(next_task)
        state += TaskSession(task_session_id=k, student_id=1, task_id=next_task,
                             solved=True, given_up=False, end=k)
        state += Student(student_id=1, credits=0)

    # after all tasks were solved, it can pick any task, but should avoid the
    # task which was just solved
    last_task = next_task
    for k in range(m, 2*m):
        next_task = recommendation.fixed_then_random(state, student_id=1)
        assert next_task != last_task
        last_task = next_task
        state += TaskSession(task_session_id=k, student_id=1, task_id=next_task,
                             solved=True, given_up=False, end=k)
        state += Student(student_id=1, credits=0)


def test_multicriteria_single_task():
    state = empty + s1 + t3
    task_id = recommendation.multicriteria(state, student_id=1, criteria=[])
    assert task_id == 3


def test_multicriteria():
    state = empty + s1 + t2 + t3
    f1 = lambda w, s, t: s + t
    f2 = lambda w, s, t: s - t
    criteria_a = [recommendation.Criterion(0.7, f1), recommendation.Criterion(0.3, f2)]
    criteria_b = [recommendation.Criterion(0.3, f1), recommendation.Criterion(0.7, f2)]
    assert recommendation.multicriteria(state, student_id=1, criteria=criteria_a) == 3
    assert recommendation.multicriteria(state, student_id=1, criteria=criteria_b) == 2
